// An interpreter for a simple script language.
//
// This language is designed to be both simple and extensible. And
// also to be written in less than an hour.


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <map>
#include <vector>
#include <malloc.h>

using namespace std;

// This stores the game script.
vector<char *> script; 

struct ltstr {
    bool operator()(const char *a, const char *b) {
        return strcmp(a, b) < 0;
    }
};

// This is a map from label to line number
map<const char *, int, ltstr> script_line_map;

// This deallocates the script. Use it if one needs to reload the
// script from scratch, for some purpose.
void script_free() {
    for (int i = 0; i < script.size(); i++) {
        free(script[i]);
    }

    script_line_map.clear();
    script.clear();
}

// Loads in the script from the named file.
void script_load(const char *fn) {
    char buf[8192];

    script_free();

    FILE *f = fopen(fn, "r");
    
    if (! f) {
        fprintf(stderr, "Could not open %s.\n", fn);
        abort();
    }

    // Add a comment to the start of the file. This makes sure that
    // lines in the script correspond to lines in the file.
    script.push_back("# This is not a line.");
    
    // While we've read a line.
    while(fgets(buf, 1024, f)) {

        // Strip CR and NL from the end of that line.
        char *c = buf;
        do {
            if (*c == '\r' || *c == '\n') *c = 0;
        } while (*c++);

        
        // Duplicates the line in memory.
        char *line = strdup(buf);

        // Adds the line to the script.
        script.push_back(line);
    }

    fclose(f);

    // Now, go in and setup the map between label and line number.
    for (int i = 0; i < script.size(); i++) {
        char *line = script[i];
        
        if (line[0] == 'L') {
            if (script_line_map[&line[1]] != 0) {
                fprintf(stderr, "Label %s exists on lines %d and %d.\n", &line[1], i, &line[1]);
                abort();
            }

            script_line_map[&line[1]] = i;
        }
    }
    
}

// Finds the line number corresponding to the supplied label. This
// reports an error if the label cannot be found.
int script_find_line(const char *label) {
    int rv = script_line_map[label];
    if (rv == 0) {
        fprintf(stderr, "Couldn't find the label %s.\n", label);
        abort();
    }

    return rv;
}

// This is the function that is responsible for interprets the
// script. Give it a label, and it runs code starting from that label
// until an R line is reached.
void script_run(const char *label) {
    int pc = script_find_line(label);

    while (pc) {
        if (pc > script.size()) {
            fprintf(stderr, "Execution fell off the edge of the world.");
            abort();
        }

        int lineno = pc;
        const char *line = script[pc];
        pc++;

        // The command that we are running.
        char cmd = line[0];

        // The argument to that command.
        const char *arg = &line[1];

        switch (cmd) {
        case 'L':
        case 0:
        case '#':
            // Ignore labels, empty lines, and comments.
            break;

        case '"':
            // Say text.
            printf("Saying: %s\n", arg);
            break;

        case '\'':
            // Narrate text.
            printf("Narrating: %s\n", arg);
            break;

        case 'S':
            // Show.
            printf("Showing: %s\n", arg);
            break;

        case 'H':
            // Hide.
            printf("Hide: %s\n", arg);
            break;

        case 'J':
            // Jump.
            pc = script_find_line(arg);
            break;

        case 'R':
            return;

        default:
            fprintf(stderr, "Unknown command on line %d: command=%c, arg=%s\n",
                    lineno, cmd, arg);
        }           
    }
}

int main(int argc, char **argv) {
    script_load("script.txt");
    script_run("start");
}
